DROP PROCEDURE IF EXISTS audit_access;
DELIMITER ||
CREATE PROCEDURE audit_access
(
 IN ob_schema TEXT,
 IN ob_name TEXT,
 IN access_priv TEXT 
)
 BEGIN
 
   IF NOT FIND_IN_SET(access_priv, 'SELECT,INSERT,UPDATE,DELETE') THEN
     SIGNAL SQLSTATE '45000'
       SET MESSAGE_TEXT = 
         'Privilege to audit must be one of:  SELECT, INSERT, UPDATE, DELETE.';
   END IF;
 
   DROP TEMPORARY TABLE IF EXISTS accounts_with_access;
   CREATE TEMPORARY TABLE accounts_with_access (
     User CHAR(32) NOT NULL,
     Host CHAR(60) NOT NULL,
     Notes TEXT,
     PRIMARY KEY (Host, User)
   );
   DROP TEMPORARY TABLE IF EXISTS accounts_with_proxy;
   CREATE TEMPORARY TABLE accounts_with_proxy (
     User CHAR(32) NOT NULL,
     Host CHAR(60) NOT NULL,
     PRIMARY KEY (Host, User)
   );
     
   -- Find users with global access:
   INSERT 
     INTO accounts_with_access (User, Host, Notes)
     SELECT User, Host, 'GLOBAL'
       FROM mysql.user
       WHERE IF(access_priv = 'SELECT', Select_priv = 'Y', 0)
         OR IF(access_priv = 'INSERT', Insert_priv = 'Y', 0)
         OR IF(access_priv = 'UPDATE', Update_priv = 'Y', 0)
         OR IF(access_priv = 'DELETE', Delete_priv = 'Y', 0)
     ON DUPLICATE KEY UPDATE Notes = CONCAT(Notes, ', GLOBAL');

-- Find users with database-specific access:
   INSERT 
     INTO accounts_with_access (User, Host, Notes)
     SELECT User, Host, 'DATABASE'
       FROM mysql.db
       WHERE ob_schema LIKE DB
         AND (IF(access_priv = 'SELECT', Select_priv = 'Y', 0)
         OR IF(access_priv = 'INSERT', Insert_priv = 'Y', 0)
         OR IF(access_priv = 'UPDATE', Update_priv = 'Y', 0)
         OR IF(access_priv = 'DELETE', Delete_priv = 'Y', 0))
     ON DUPLICATE KEY UPDATE Notes = CONCAT(Notes, ', DATABASE');

-- Find users with table-specific access:
   INSERT
     INTO accounts_with_access (User, Host, Notes)
     SELECT DISTINCT User, Host, 'TABLE'
       FROM mysql.tables_priv
       WHERE ob_schema = Db
         AND ob_name = Table_name
         AND FIND_IN_SET(access_priv, Table_priv)
     ON DUPLICATE KEY UPDATE Notes = CONCAT(Notes, ', TABLE');

-- Find users with column-specific access:
   INSERT
     INTO accounts_with_access (User, Host, Notes)
     SELECT DISTINCT User, Host, 'COLUMN'
       FROM mysql.columns_priv
       WHERE Db = ob_schema
         AND Table_name = ob_name
         AND FIND_IN_SET(access_priv, Column_priv)
     ON DUPLICATE KEY UPDATE Notes = CONCAT(Notes, ', COLUMN');

-- Find users with proxy-based access.
-- First, find the users who can proxy users
-- we know have access:
   INSERT
     INTO accounts_with_proxy (User, Host)
     SELECT DISTINCT pp.User, pp.Host
       FROM mysql.proxies_priv pp
       JOIN accounts_with_access awa
         ON (awa.Host = pp.Proxied_Host AND awa.User = pp.Proxied_User);
 
-- Then add these users to the list:
   INSERT
     INTO accounts_with_access (User, Host, Notes)
     SELECT DISTINCT awp.User, awp.Host, 'PROXY'
       FROM accounts_with_proxy awp
     ON DUPLICATE KEY UPDATE Notes = CONCAT(Notes, ', PROXY');
     
   SELECT * FROM accounts_with_access;
   DROP TEMPORARY TABLE IF EXISTS accounts_with_access;
   DROP TEMPORARY TABLE IF EXISTS accounts_with_proxy;
 END||
 DELIMITER ;
 
 
 CREATE DATABASE IF NOT EXISTS access_audit_test;
 CREATE TABLE IF NOT EXISTS access_audit_test.t1 (a INT, b INT);
 
 DROP USER IF EXISTS at1@localhost;
 DROP USER IF EXISTS at2@localhost;
 DROP USER IF EXISTS at3@localhost;
 DROP USER IF EXISTS at4@localhost;
 DROP USER IF EXISTS at5@localhost;
 DROP USER IF EXISTS at6@localhost;

 CREATE USER at1@localhost, at2@localhost, at3@localhost, at4@localhost, at5@localhost, at6@localhost;
 GRANT SELECT ON *.* TO at1@localhost;
 GRANT SELECT ON access_audit_test.* TO at2@localhost;
 GRANT SELECT ON access_audit_test.t1 TO at3@localhost;
 GRANT SELECT (a) ON access_audit_test.t1 TO at4@localhost;
 GRANT PROXY ON at1@localhost TO at5@localhost;
 GRANT SELECT ON `a%`.* TO at6@localhost;
 GRANT SELECT ON *.* TO at6@localhost;
 
 CALL audit_access('access_audit_test', 't1', 'SELECT');
 